#! /usr/local/bin/apl --script
⍝ ********************************************************************
⍝ install_packages.apl Workspace to install library in package manager.
⍝ Copyright (C) 2020 Bill Daly

⍝ This program is free software: you can redistribute it and/or modify
⍝ it under the terms of the GNU General Public License as published by
⍝ the Free Software Foundation, either version 3 of the License, or
⍝ (at your option) any later version.

⍝ This program is distributed in the hope that it will be useful,
⍝ but WITHOUT ANY WARRANTY; without even the implied warranty of
⍝ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
⍝ GNU General Public License for more details.

⍝ You should have received a copy of the GNU General Public License
⍝ along with this program.  If not, see <http://www.gnu.org/licenses/>.

⍝ ********************************************************************

)copy 1 utl
)copy 5 FILE_IO
)copy 1 cfg
)copy 1 preferences
)copy 1 ancillary/pkg_manager_lib

∇ old ins∆copyFile new;old_data;newh;oldh;sink;stat
  ⍝ Function to copy a file, old and new are full paths to the files.
  utl∆es (0>oldh←'r' FIO∆fopen old)/'Error reading ',old
  stat←FIO∆fstat oldh
  old_data←stat[8] FIO∆fread oldh
  sink←FIO∆fclose oldh
  utl∆es (0>newh←'w' FIO∆fopen new)/'Error writing ',new
  old_data FIO∆fwrite newh
  sink←FIO∆fclose newh
∇

∇ src ins∆copy_test_data target;files;lb;ix
  ⍝ Function to copy test data to the package manager.
  →(aplpu∆file_exists target)/copy
  FIO∆mkdir_777 target
copy:
  files ← FIO∆files_in_directory src
  lb←((1↑⍴files)⍴st),ed
  ix←1
st:
  (src,'/',ix⊃files) ins∆copyFile target,'/',ix⊃files
  →lb[ix←ix+1]
ed:
  ∇

∇ home←ins∆library∆home ;pref;wsid;lib
  ⍝ Function returns a path to apl-library.
  prefs←pref∆current pref∆find_prefs
  wsid←' ' utl∆split ,⊃⍎')wsid'
  →(2=⍴wsid)/libZero
  utl∆es (3≠⍴wsid)/'UNABLE TO DETERMINE CURRENT LIBRARY'
  lib←2⊃wsid
  →(~prefs lex∆haskey 'LIBREF-',lib)/NoPrefs
  home←prefs lex∆lookup 'LIBREF-',lib
  →proof
libZero:
  ⍎(prefs lex∆haskey 'LIBREF-0')/'home←prefs lex∆lookup ''LIBREF-0''◊→proof'
  home←(2⊃,⎕env 'HOME'),'/workspaces'
  →proof
NoPrefs:
  home←(2⊃,⎕env 'HOME'),'/wslib',lib
  →proof
proof:
  utl∆es (~'apl-pkg-template' utl∆stringMember FIO∆files_in_directory home)/'APL-LIBRARY NOT FOUND.'
  →0
∇

∇ libs ins∆copyToPkgManager instance;libHome;pkgHome;txt;instanceDir
  ⍝ Function copies files to package manager. libs is two items
  ⍝  1. library home
  ⍝  2. package manager home
  (libHome pkgHome)←libs
  ⍝ Create package manager subdirectory
  →(instance utl∆stringMember FIO∆files_in_directory pkgHome)/copyFiles
  511 FIO∆mkdir pkgHome,'/',instance
  ⍝ Copy files to package manag
copyFiles:
  ⍝ Create documentation
  txt←FIO∆pipefrom 'makeinfo --plaintext --output=',pkgHome,'/',instance,'/',instance,'.txt ',libHome,'/doc/',instance,'.texi'
  ⍝txt utf8∆write pkgHome,'/',instance,'/',instance,'.txt'
  (libHome,'/',instance,'.apl') ins∆copyFile pkgHome,'/',instance,'/',instance,'.apl'
  (libHome,'/apl-pkg-template/',instance,'/_control_.apl') ins∆copyFile pkgHome,'/',instance,'/_control_.apl'
  (libHome,'/apl-pkg-template/',instance,'/_metadata_') ins∆copyFile pkgHome,'/',instance,'/_metadata_'
∇

∇libHome ins∆unitTestToPkgManager pkgHome;testWS;template;lbl;ix;fname
  ⍝ Functio to Build the unit-test package
  target←pkgHome,'/unit_test'
  template←libHome,'/apl-pkg-template/unit_test'
  →('unit_test' utl∆stringMember FIO∆files_in_directory pkgHome)/copyFiles
  511 FIO∆mkdir target
copyFiles:
  (template,'/_control_.apl') ins∆copyFile target,'/_control_.apl'
  (template,'/_metadata_') ins∆copyFile target,'/_metadata_'
   testWS←FIO∆files_in_directory libHome
  testWS←({∧/'test.apl'=¯8↑⍵}¨testWS)/testWS
  ix←1
  lbl←((⍴testWS)⍴st),ed
st:
  fname←ix⊃testWS
  (libHome,'/',fname) ins∆copyFile target,'/',fname
  →lbl[ix←ix+1]
ed:
  (libHome,'/test_data') ins∆copy_test_data pkgHome,'/unit_test/test_data'
∇


∇ ins∆main;pkgTemplate;pkgHome;libHome
  ⍝ Function adds apl-library workspaces to the apl-packages system.
  pkgHome←aplpu∆pkgMgr∆home
  libHome←ins∆library∆home
  libHome ins∆unitTestToPkgManager pkgHome
  pkgTemplate←FIO∆files_in_directory libHome,'/apl-pkg-template'
  pkgTemplate←(~{∧/'Makefile'=8↑⍵}¨pkgTemplate)/pkgTemplate
  pkgTemplate←(~{∧/'unit_test'=9↑⍵}¨pkgTemplate)/pkgTemplate
  (⊂libHome pkgHome) ins∆copyToPkgManager ¨ pkgTemplate
  libHome ins∆unitTestToPkgManager pkgHome
∇

⍝ins∆main
⍝)off
