⍝ This is a package control file.

⍝ Decline to load this file unless the package manager is present.
⍎(0=⎕nc'pkg∆manager')/'''Load using the APL Package Manager.'' ◊ →'

⍝ The following information applies to the APL Package Manager release
⍝ named "Percy".
⍝
⍝ A package consists of a directory which in turn contains files or
⍝ other directories. The files, with the exception of the control and
⍝ metadata files, may contain anything. The directories may be used to
⍝ organize other files and directories in any manner. The only requirements
⍝ of the packaging scheme are the top-level directory (nominally named for
⍝ the package) and the control and metadata files.
⍝
⍝ For purposes of transport and distribution, a package may be bundled
⍝ in a .zip file. The zip file should bear the name of the package
⍝ directory.
⍝
⍝ A package may be installed via other means, such as from a version
⍝ control repository such as used by svn or git. The presence of version
⍝ control data in the package directory shall not influence the function
⍝ of the package manager.
⍝ 
⍝ The package control file is just a Unicode representation of a valid
⍝ APL program. The control file may be complete in and of itself or
⍝ may run code to construct, load and install the package using additional
⍝ files and tools.
⍝
⍝ A package may depend on other packages, but must not access or
⍝ create files in other package directories.
⍝
⍝ It's perfectly fine for a package to read or install files in
⍝ locations other than in package directories. Of course, you must
⍝ qualify external locations based upon platform probes.
⍝
⍝ The following table lists the API functions that you may use to
⍝ facilitate loading of a package. Refer to API.md for detailed
⍝ descriptions.
⍝
⍝ | Function            | Use                                   | Phase |
⍝ | ------------------- | ------------------------------------- | ----- |
⍝ | pkg∆manager         | package manager release               | Q     |
⍝ | pkg∆platform_family | host platform kind                    | R     |
⍝ | pkg∆os_type         | host OS identification                | R     |
⍝ | pkg∆os_distribution | host OS identification                | R     |
⍝ | pkg∆os_version      | host OS identification                | R     |
⍝ | pkg∆apl_type        | APL system identification             | R     |
⍝ | pkg∆apl_version     | APL system identification             | R     |
⍝ | pkg∆shell_type      | host shell identification             | R     |
⍝ | pkg∆shell_version   | host shell identification             | R     |
⍝ | pkg∆case            | select alternatives functions         | L     |
⍝ | pkg∆compare_version | test relation between version numbers | L     |
⍝ | pkg∆shell           | run host shell command                | L     |
⍝ | pkg∆copy            | copy APL code into workspace          | L     |
⍝ | pkg∆file            | query host filesystem objects         | L     |
⍝ | pkg∆alias           | alias a defined function or operator  | L [1] |
⍝
⍝ Phases
⍝ ------
⍝ Q: Query - Determine the presence of the package manager via
⍝   `0≠⎕nc 'pkg∆manager'`.
⍝ L: Load-time - These functions may only be called during package
⍝    loading. [1] The pkg∆alias function must not be invoked when
⍝    loading a package from a manifest.
⍝ R: Run-time - These functions may be called at package run time if
⍝    guarded by a query for package manager presence. These functions
⍝    may also be called during package loading.

∇ test∆dispatcher prefix;nl;lb;ix;width
  ⍝ Function to dispatch test functions
  width←⍴prefix←,prefix
  nl←(nl[;⍳width]∧.=prefix)⌿nl←⎕nl 3
  ix←1
  lb←((1↑⍴nl)⍴st),ed
st:⍎nl[ix;]
  →lb[ix←ix+1]
ed:
∇

∇dir ← test_data;config
  ⍝ Function returns the path to the package directory.
  config← cfg∆parse_file (2⊃,⎕env'HOME'),'/.apl-pkg'
  dir←(config lex∆lookup 'DEFAULT') lex∆lookup 'packages_directory'
  dir←dir,'/unit_test/test_data'
∇

∇test∆menu test∆menu∆items;resp;count;item
  ⍝ Function to display list of packages to test and execute requested
  ⍝ package tests.
  ⍞←count←1↑⍴test∆menu∆items
  ⍞←⎕tc[3]
tp:				⍝ Done
  ('550 '⍕(count,1)⍴⍳count),⊃test∆menu∆items
  resp←prompt 'What file should I execute?  '
  ⍞←resp,⎕tc[3]
  ⍎(utl∆numberp resp)/'→(qu,tp,dn)[3⌊⎕io+resp]'
  →(sresp,nresp)[1+utl∆numberis resp]
sresp:
  item←test∆menu∆items utl∆listSearch resp
  →ex
nresp:
  item←⍎resp
  →((item>count)∨item≤0)/nerr
ex:
  ⍎item⊃test∆menu∆items
  →tp
qu:				⍝ Quit
  →0
dn:				⍝ Done
  →0
nerr:
  ⍞←'Your response should be between 1 and ',(⍕count),'.'
  →tp
∇
test∆menu∆items←⍬
test∆menu∆items←test∆menu∆items,⊂'test∆date'
pkg∆copy 'date_test.apl'
test∆menu∆items←test∆menu∆items,⊂'test∆dom'
pkg∆copy 'dom_test.apl'
test∆menu∆items←test∆menu∆items,⊂'test∆exim'
pkg∆copy 'exim_test.apl'
test∆menu∆items←test∆menu∆items,⊂'test∆finance'
pkg∆copy 'finance_test.apl'
test∆menu∆items←test∆menu∆items,⊂'test∆fmt'
pkg∆copy 'fmt_test.apl'
test∆menu∆items←test∆menu∆items,⊂'test∆lex'
pkg∆copy 'lex_test.apl'
test∆menu∆items←test∆menu∆items,⊂'test∆stat'
pkg∆copy 'stat_test.apl'
test∆menu∆items←test∆menu∆items,⊂'test∆utf8'
pkg∆copy 'utf8_test.apl'
test∆menu∆items←test∆menu∆items,⊂'test∆utl'
pkg∆copy 'utl_test.apl'
test∆menu∆items←test∆menu∆items,⊂'test∆cfg'
pkg∆copy 'cfg_test.apl'





