#! /usr/local/bin/apl --script
⍝ ********************************************************************
⍝ cl_test, unit testing form workspace cl
⍝ Copyright (C) 2019 Bill Daly

⍝ This program is free software: you can redistribute it and/or modify
⍝ it under the terms of the GNU General Public License as published by
⍝ the Free Software Foundation, either version 3 of the License, or
⍝ (at your option) any later version.

⍝ This program is distributed in the hope that it will be useful,
⍝ but WITHOUT ANY WARRANTY; without even the implied warranty of
⍝ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
⍝ GNU General Public License for more details.

⍝ You should have received a copy of the GNU General Public License
⍝ along with this program.  If not, see <http://www.gnu.org/licenses/>.

⍝ ********************************************************************

)copy 1 cl
)copy 5 FILE_IO

⍝ cl is a lexicon implemented as a apl component file (5
⍝ APLComponentFiles). The first task is connecting to a postgresql
⍝ server and the second creating a database.

)copy 1 assert
)copy 1 import
)copy 1 find_test_data

⍝ Your SQL administrator must set you up as a user.  My administrator
⍝ always creates a database so that I can make a connection.  You'll
⍝ need both.

∇test_cl_setup;sqlh
  ⍝ Function to set up the test environment
  test_cl_type ← 'sqlite'
  test_cl_dir←test_data
  test_cl_spec←test_cl_dir,'/test_cl.sqlite'
  test_cash_flow_I←import∆file test_cl_dir,'/cash_flow_I.txt'
  test_f500←import∆file test_cl_dir,'/f500.txt'
  test_fifthq←import∆file test_cl_dir,'/inv-fifthQ.txt'
  test_websites←import∆file test_cl_dir,'/inv_f500_websites.txt'
∇

∇rs←test_cl_connect;sink
  ⍝ Function attempts to connect to the database; It will create a new
  ⍝ database if that fails
  test_cl_type cl∆open_db test_cl_spec
  →(0<SQL∆Tables _CF_DB)/0
  sink←FIO∆pipefrom 'rm ',test_cl_spec
  test_cl_type CF_DBCREATE test_cl_spec
∇


∇test_cl_cleanup
  ⍝ Function to close the component file and the database connection
  cl∆close cfh
  cl∆erase 'cl_test01'
  cl∆close_db
∇

∇test∆cl
  assert∆setup 'test_cl_setup'
  '2=⎕nc ''_CF_DB''' assert∆nil∆toScreen 'test_cl_type cl∆open_db test_cl_spec'
  assert∆true assert∆toScreen 'utl∆numberp cfh←cl∆init ''cl_test01'''
  'cfh cl∆haskey ''cash_flow_I''' assert∆nil∆toScreen 'cfh cl∆assign ''cash_flow_I'' test_cash_flow_I'
  'cfh cl∆haskey ''fifthq''' assert∆nil∆toScreen 'cfh cl∆assign ''fifthq'' test_fifthq'
  'cfh cl∆haskey ''websites''' assert∆nil∆toScreen 'cfh cl∆assign ''websites'' test_websites'
  'cfh cl∆haskey ''f500''' assert∆nil∆toScreen 'cfh cl∆assign ''f500'' test_f500'
  test_cash_flow_I assert∆toScreen 'cfh cl∆lookup ''cash_flow_I'''
  test_fifthq assert∆toScreen 'cfh cl∆lookup ''fifthq'''
  test_websites assert∆toScreen 'cfh cl∆lookup ''websites'''
  test_f500 assert∆toScreen 'cfh cl∆lookup ''f500'''
  assert∆cleanUp 'cl∆close cfh' 'cl∆erase ''cl_test01''' 'cl∆close_db'
∇


⍝)off
